from __future__ import print_function, division

__author__ = 'amrit'

import sys

#sys.dont_write_bytecode = True
from sklearn.tree import DecisionTreeClassifier
from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import SVC
from sklearn.metrics import f1_score, accuracy_score,confusion_matrix
import numpy as np
import pandas as pd
#from .FFT import FFT

metrics=['accuracy','recall','precision','false_alarm']

metrics_dic={'accuracy':-2,'recall':-6,'precision':-7,'false_alarm':-4}

def DT(k,train_data,train_labels,test_data,test_labels, metric):

    model = DecisionTreeClassifier(**k)
    model.fit(train_data, train_labels)
    prediction=model.predict(test_data)
    dic = {}

    for i in metrics:
        dic[i] = round(evaluation(i, prediction, test_labels),3)

    return dic[metric], [dic, model.feature_importances_]

def RF(k,train_data,train_labels,test_data,test_labels, metric):
    model = RandomForestClassifier(**k)
    model.fit(train_data, train_labels)
    prediction = model.predict(test_data)
    dic = {}
    for i in metrics:
        dic[i] = round(evaluation(i, prediction, test_labels),3)
    return dic[metric], [dic, model.feature_importances_]


def SVM(k,train_data,train_labels,test_data,test_labels, metric):
    from sklearn.preprocessing import MinMaxScaler
    scaling = MinMaxScaler(feature_range=(-1, 1)).fit(train_data)
    train_data = scaling.transform(train_data)
    test_data = scaling.transform(test_data)
    model = SVC(cache_size=20000,**k)
    model.fit(train_data, train_labels)
    #print(model.coef_)
    prediction = model.predict(test_data)
    dic = {}
    print(prediction)
    for i in metrics:
        dic[i] = round(evaluation(i, prediction, test_labels),3)
    return dic[metric], [dic, []]


def FFT1(k,train_data,train_labels,test_data,test_labels, metric):
    dic={}
    dic1={}
    for i in metrics:
        fft = FFT(max_level=5)
        fft.criteria=i
        #fft.print_enabled=True
        train_labels=np.reshape(train_labels,(-1,1))
        test_labels = np.reshape(test_labels, (-1, 1))

        training=np.hstack((train_data, train_labels))
        testing = np.hstack((test_data, test_labels))
        training_df = pd.DataFrame(training)
        testing_df = pd.DataFrame(testing)
        training_df.rename(columns={training_df.columns[-1]: "bug"},inplace=True)
        testing_df.rename(columns={testing_df.columns[-1]: "bug"},inplace=True)

        fft.target = "bug"
        fft.train, fft.test = training_df, testing_df
        fft.build_trees()  # build and get performance on TEST data
        t_id = fft.find_best_tree()  # find the best tree on TRAIN data
        fft.eval_tree(t_id)  # eval all the trees on TEST data

        description=fft.print_tree(t_id)

        dic[i]=fft.performance_on_test[t_id][metrics_dic[i]]
        dic1[i]=description
    return dic[metric], [dic, dic1]


def evaluation(measure, prediction, test_labels, class_target=1):
    confu = confusion_matrix(test_labels, prediction)
    fp = confu.sum(axis=0) - np.diag(confu)
    fn = confu.sum(axis=1) - np.diag(confu)
    tp = np.diag(confu)
    tn = confu.sum() - (fp + fn + tp)
    if measure == "accuracy":
        return accuracy_score(test_labels, prediction)
    if measure == "recall":
        recall = 0
        if class_target == -1:
            for m in range(len(tp)):
                if tp[m] != 0 and (tp[m] + fn[m]) != 0:
                    recall += float(tp[m]) / (tp[m] + fn[m])
            return recall / len(tp)
        else:
            if tp[class_target] != 0 and (tp[class_target] + fn[class_target]) != 0:
                return float(tp[class_target]) / (tp[class_target] + fn[class_target])
            else:
                return 0.0

    if measure == "precision":
        precision = 0
        if class_target == -1:
            for m in range(len(tp)):
                if tp[m] != 0 and (tp[m] + fp[m]) != 0:
                    precision += float(tp[m]) / (tp[m] + fp[m])
            return precision / len(tp)
        else:
            if tp[class_target] != 0 and (tp[class_target] + fp[class_target]) != 0:
                return float(tp[class_target]) / (tp[class_target] + fp[class_target])
            else:
                return 0.0
    if measure == "false_alarm":
        fals=0
        if class_target==-1:
            for m in range(len(fp)):
                if fp[m] != 0 and (fp[m] + tn[m]) !=0 :
                    fals+=float(fp[m])/(fp[m]+tn[m])
            return fals/len(fp)
        else:
            if fp[class_target] != 0 and (fp[class_target] + tn[class_target]) != 0:
                return float(fp[class_target]) / (fp[class_target] + tn[class_target])
            else:
                return 0.0

    if measure == "f1":
        if class_target==-1:
            return f1_score(test_labels, prediction, average='macro')
        else:
            return f1_score(test_labels, prediction, pos_label=class_target, average='binary')


from __future__ import print_function, division

__author__ = 'amrit'

import sys
from .demo import cmd
#sys.dont_write_bytecode = True
from .DE import DE
from .ldavem import *
from collections import OrderedDict
import os
from collections import Counter
from .featurization import *
from .ML import DT, SVM, RF, FFT1
from sklearn.model_selection import StratifiedKFold
import pickle
import time

learners=[main]
learners_para_dic=[OrderedDict([("n_components",10),("doc_topic_prior",0.1), ("topic_word_prior",0.01)])]
learners_para_bounds=[[(10,100), (0.1,1), (0.01,1)]]
learners_para_categories=[[ "integer", "continuous", "continuous"]]
ROOT=os.getcwd()
files=["pitsA", "pitsB", "pitsC", "pitsD", "pitsE", "pitsF"]
MLS=[DT,RF, SVM,  FFT1]
MLS_para_dic = {'DT':OrderedDict([("min_samples_split", 2), ("min_impurity_decrease", 0.0), ("max_depth", None),
                             ("min_samples_leaf", 1)]), 
                'RF': OrderedDict([("min_samples_split", 2),("max_leaf_nodes", None), ("min_samples_leaf", 1),
                                   ("min_impurity_decrease", 0.0),("n_estimators", 10)]),
                'SVM': OrderedDict([("C", 1.0), ("kernel", 'linear'),
                             ("degree", 3)]), 
                'FFT1':OrderedDict()}

metrics=['accuracy','recall','precision','false_alarm']
def readfile1(filename=''):
    dict = []
    labels=[]
    with open(filename, 'r') as f:
        for doc in f.readlines():
            try:
                row = doc.lower().split(">>>")
                dict.append(row[0].strip())
                labels.append(row[1].strip())
            except:
                pass
    count=Counter(labels)
    import operator
    key = max(count.iteritems(), key=operator.itemgetter(1))[0]
    labels=map(lambda x: 1 if x == key else 0, labels)
    return np.array(dict), np.array(labels)

def _test(res='',_MLS = [DT, RF, SVM,FFT1],file_path = ''):
    seed(1)
    np.random.seed(1)
    #path=ROOT+"/data/preprocessed/"+res+".txt"
    path = file_path + res + ".txt"
    raw_data,labels=readfile1(path)
    temp={}
    MLS = _MLS
    for i in range(5):
        ranges=range(len(labels))
        shuffle(ranges)
        raw_data=raw_data[ranges]
        labels=labels[ranges]
        #print(raw_data)
        start_time=time.time()
        de = DE(Goal="Max", GEN=5, NP=10,termination="Early")
        v, _ = de.solve(learners[0], OrderedDict(learners_para_dic[0]),
                        learners_para_bounds[0], learners_para_categories[0],
                        file=res, term=7, data_samples=raw_data)
        corpus,_=LDA_(raw_data,**v.ind)
        end_time = time.time()-start_time

        skf = StratifiedKFold(n_splits=5)
        for train_index, test_index in skf.split(corpus, labels):
            train_data, train_labels = corpus[train_index], labels[train_index]
            test_data, test_labels = corpus[test_index], labels[test_index]

            for j, le in enumerate(MLS):
                if le.__name__ not in temp:
                    temp[le.__name__]={}
                start_time1=time.time()
                _,val=MLS[j](MLS_para_dic[le.__name__], train_data, train_labels, test_data, test_labels, 'recall')
                end_time1=time.time()-start_time1
                for m in metrics:
                    if m not in temp[le.__name__]:
                        temp[le.__name__][m]=[]
                    temp[le.__name__][m].append(val[0][m])
                if 'times' not in temp[le.__name__]:
                    temp[le.__name__]['times']=[]
                else:
                    temp[le.__name__]['times'].append(end_time1+end_time)
                if 'features' not in temp[le.__name__]:
                    temp[le.__name__]['features'] = []
                else:
                    temp[le.__name__]['features'].append(val[1])

    return temp

if __name__ == '__main__':
    eval(cmd ())



#!/usr/bin/env python2
# -*- coding: utf-8 -*-
"""
Created on Sat Sep 15 13:36:17 2018

@author: suvodeepmajumder
"""

from __future__ import print_function, division
import sys
from .demo import cmd
#sys.dont_write_bytecode = True
from .DE import DE
from .ldavem import *
from collections import OrderedDict
import os
from collections import Counter
from .featurization import *
from .ML import SVM,DT,RF
from sklearn.model_selection import StratifiedKFold
import pickle
import time
import numpy as np



lda=[main]
lds_para_dic=[OrderedDict([("n_components",10),("doc_topic_prior",0.1), ("topic_word_prior",0.01)])]
lda_para_bounds=[[(10,100), (0.1,1), (0.01,1)]]
lda_para_categories=[[ "integer", "continuous", "continuous"]]


learners=[SVM,DT,RF]
learners_para_dic={'SVM':OrderedDict([("C", 1.0), ("kernel", 'linear'), ("degree", 3)]),
                    'DT': OrderedDict([("criterion",'entropy'), ("splitter", 'best'), ("min_samples_split", 2), 
                                 ("min_samples_leaf", 1)]),
                    'RF':OrderedDict([("n_estimators", 10), ("criterion",'entropy'), ("min_samples_split", 2), 
                                 ("min_samples_leaf", 1)])}
learners_para_bounds={'SVM': [(0.1,100), ("linear","poly","rbf","sigmoid"), (1,20)],
                       'DT': [("gini","entropy"),("best","random"),(2,50),(1,20)],
                       'RF': [(10,100),("gini","entropy"),(2,50),(1,20)]}
learners_para_categories={'SVM': ["continuous", "categorical", "integer"],
                          'DT': ["categorical","categorical","integer","integer"],
                          'RF': ["integer","categorical","integer","integer"]}
ROOT=os.getcwd()
files=["pitsA", "pitsB", "pitsC", "pitsD", "pitsE", "pitsF"]
features=[LDA_]
topics = ['10']
metrics=['accuracy','recall','precision','false_alarm']

def readfile1(filename=''):
    dict = []
    labels=[]
    with open(filename, 'r') as f:
        for doc in f.readlines():
            try:
                row = doc.lower().split(">>>")
                dict.append(row[0].strip())
                labels.append(row[1].strip())
            except:
                pass
    count=Counter(labels)
    import operator
    key = max(count.iteritems(), key=operator.itemgetter(1))[0]
    labels=map(lambda x: 1 if x == key else 0, labels)
    return np.array(dict), np.array(labels)

def _test(res='',_learners = [SVM,DT,RF],file_path = ''):
    learners = []
    learners = _learners
#    print(_learners)
#    for _learner in _learners:
#        if _learner == 'SVM':
#            learners.append(SVM)
#        elif _learner == 'DT':
#            learners.append(DT)
#        elif _learner == 'RF':
#            learners.append(RF)
#    print(learners)
    seed(1)
    np.random.seed(1)
    path = file_path + res + ".txt"
    raw_data, labels = readfile1(path)
    temp={}
    for i in range(2):
        print('Starting iteration ',i)
        ranges=range(len(labels))
        shuffle(ranges)
        raw_data=raw_data[ranges]
        labels=labels[ranges]
        for fea in features:
            start_time = time.time()
            corpus,_=fea(raw_data)
            end_time = time.time() - start_time

#                    start_time=time.time()
#                    de = DE(Goal="Max", GEN=5, NP=10,termination="Early")
#                    v, _ = de.solve(lda[0], OrderedDict(lds_para_dic[0]),
#                                    lda_para_bounds[0],
#                                    lda_para_categories[0],file=res, term=7, data_samples=raw_data)
#                    corpus,_=LDA_(raw_data,**v.ind)
#                    end_time = time.time()-start_time

            skf = StratifiedKFold(n_splits=2)
            for train_index, test_index in skf.split(corpus, labels):
                print('Starting fold :')
                train_data, train_labels = corpus[train_index], labels[train_index]
                test_data, test_labels = corpus[test_index], labels[test_index]
                
                for DE_train_index, validation_index in skf.split(train_data, train_labels):
                    DE_train_data, DE_train_labels = train_data[DE_train_index], train_labels[DE_train_index]
                    DE_validation_data, DE_validation_labels = corpus[validation_index], labels[validation_index]
                    for j, le in enumerate(learners):
                        print('Start DE')
#                        print(learners[j], OrderedDict(learners_para_dic[le.__name__]))
                        ML_de = DE(Goal="Max", GEN=5, NP=10,termination="Early")
                        print(le,le.__name__,OrderedDict(learners_para_dic[le.__name__]))
                        mlv, _ = ML_de.solve(le, OrderedDict(learners_para_dic[le.__name__]),
                                learners_para_bounds[le.__name__],
                                learners_para_categories[le.__name__],
                                train_data = DE_train_data,train_labels = DE_train_labels, 
                                test_data = DE_validation_data, test_labels = DE_validation_labels, metric = 'recall')
                        print('End DE')
                        if le.__name__ not in temp:
                            temp[le.__name__]={}
                        start_time1=time.time()
                        _,val=le(mlv.ind, train_data, train_labels, test_data, test_labels, 'recall')
                        end_time1=time.time()-start_time1

                        if 'times' not in temp[le.__name__]:
                            temp[le.__name__]['times']=[]
                        else:
                            temp[le.__name__]['times'].append(end_time1+end_time)
                        if 'metrices' not in temp[le.__name__]:
                            temp[le.__name__]['metrices'] = []
                        else:
                            temp[le.__name__]['metrices'].append(val[0])

#    print(temp)
    return temp

if __name__ == '__main__':
    eval(cmd ())



from __future__ import division

__author__ = 'amrit'

import sys

#sys.dont_write_bytecode = True

def atom(x):
    try:
        return int(x)
    except ValueError:
        try:
            return float(x)
        except ValueError:
            return x


def cmd(com="demo('-h')"):
    "Convert command line to a function call."
    if len(sys.argv) < 2: return com

    def strp(x): return isinstance(x, basestring)

    def wrap(x): return "'%s'" % x if strp(x) else str(x)

    words = map(wrap, map(atom, sys.argv[2:]))
    return sys.argv[1] + '(' + ','.join(words) + ')'


def demo(f=None, cache=[]):
    def doc(d):
        return '# ' + d.__doc__ if d.__doc__ else ""

    if f == '-h':
        print('# sample demos')
        for n, d in enumerate(cache):
            print('%3s) ' % (n + 1), d.func_name, doc(d))
    elif f:
        cache.append(f)
    else:
        s = '|' + '=' * 40 + '\n'
        for d in cache:
            print('\n==|', d.func_name, s, doc(d), d())
    return f


def test(f=None, cache=[]):
    if f:
        cache += [f]
        return f
    ok = no = 0
    for t in cache:
        print('#', t.func_name, t.__doc__ or '')
        prefix, n, found = None, 0, t() or []
        while found:
            this, that = found.pop(0), found.pop(0)
            if this == that:
                ok, n, prefix = ok + 1, n + 1, 'CORRECT:'
            else:
                no, n, prefix = no + 1, n + 1, 'WRONG  :'
            print(prefix, t.func_name, 'test', n)
    if ok + no:
        print('\n# Final score: %s/%s = %s%% CORRECT' \
              % (ok, (ok + no), int(100 * ok / (ok + no))))


@test
def tested():
    return [True, True,  # should pass
            False, True,  # should fail
            1, 2 / 2]  # should pass


@demo
def demoed(show=1):
    "Sample demo."
    print(show / 2)


@demo
def tests():
    "Run all the test cases."
    test()



from __future__ import print_function, division

__author__ = 'amrit'

import sys
from demo import cmd
#sys.dont_write_bytecode = True
from DE import DE
from ldavem import *
from collections import OrderedDict
import os
from collections import Counter
from featurization import *
from ML import SVM
from sklearn.model_selection import StratifiedKFold
import pickle
import time
import pandas as pd

lda=[main]
lds_para_dic=[OrderedDict([("n_components",10),("doc_topic_prior",0.1), ("topic_word_prior",0.01)])]
lda_para_bounds=[[(10,100), (0.1,1), (0.01,1)]]
lda_para_categories=[[ "integer", "continuous", "continuous"]]


learners=[SVM]
learners_para_dic=[OrderedDict([("C", 1.0), ("kernel", 'linear'), ("degree", 3)])]
learners_para_bounds=[[(0.1,100), ("linear","poly","rbf","sigmoid"), (1,20)]]
learners_para_categories=[["continuous", "categorical", "integer"]]
ROOT=os.getcwd()
files=["pitsA", "pitsB", "pitsC", "pitsD", "pitsE", "pitsF"]
features=[LDA_]
topics = ['10']
metrics=['accuracy','recall','precision','false_alarm']

def readfile1(filename=''):
    dict = []
    labels=[]
    with open(filename, 'r') as f:
        for doc in f.readlines():
            try:
                row = doc.lower().split(">>>")
                dict.append(row[0].strip())
                labels.append(row[1].strip())
            except:
                pass
    count=Counter(labels)
    import operator
    key = max(count.iteritems(), key=operator.itemgetter(1))[0]
    labels=map(lambda x: 1 if x == key else 0, labels)
    return np.array(dict), np.array(labels)

def _test(res='',file_path = ''):
    seed(1)
    np.random.seed(1)
    path = file_path + res + ".txt"
    raw_data, labels = readfile1(path)
    temp={}

#    for m in metrics:
#    if m!='false_alarm':
#    if m not in temp:
#        temp[m] = {}
    for i in range(2):
        print('Starting iteration ',i)
        ranges=range(len(labels))
        shuffle(ranges)
        raw_data=raw_data[ranges]
        labels=labels[ranges]
        for fea in features:
            start_time = time.time()
            corpus,_=fea(raw_data)
            end_time = time.time() - start_time

#                    start_time=time.time()
#                    de = DE(Goal="Max", GEN=5, NP=10,termination="Early")
#                    v, _ = de.solve(lda[0], OrderedDict(lds_para_dic[0]),
#                                    lda_para_bounds[0],
#                                    lda_para_categories[0],file=res, term=7, data_samples=raw_data)
#                    corpus,_=LDA_(raw_data,**v.ind)
#                    end_time = time.time()-start_time

            skf = StratifiedKFold(n_splits=2)
            for train_index, test_index in skf.split(corpus, labels):
                print('Starting fold :')
                train_data, train_labels = corpus[train_index], labels[train_index]
                test_data, test_labels = corpus[test_index], labels[test_index]
                
                for DE_train_index, validation_index in skf.split(train_data, train_labels):
                    DE_train_data, DE_train_labels = train_data[DE_train_index], train_labels[DE_train_index]
                    DE_validation_data, DE_validation_labels = corpus[validation_index], labels[validation_index]
                    print('Start DE')
                    ML_de = DE(Goal="Max", GEN=5, NP=10,termination="Early")
                    mlv, _ = ML_de.solve(learners[0], OrderedDict(learners_para_dic[0]),
                            learners_para_bounds[0],
                            learners_para_categories[0],
                            train_data = DE_train_data,train_labels = DE_train_labels, 
                            test_data = DE_validation_data, test_labels = DE_validation_labels, metric = 'recall')
#                            print(mlv.ind)
                    print('End DE')
                    for j, le in enumerate(learners):
                        if le.__name__ not in temp:
                            temp[le.__name__]={}
                        start_time1=time.time()
                        _,val=learners[0](mlv.ind, train_data, train_labels, test_data, test_labels, 'recall')
                        end_time1=time.time()-start_time1
                        print(val)

                        for m in metrics:
                            if m not in temp[le.__name__]:
                                temp[le.__name__][m]=[]
                        if 'times' not in temp[le.__name__]:
                            temp[le.__name__]['times']=[]
                        else:
                            temp[le.__name__]['times'].append(end_time1+end_time)
                        if 'metrices' not in temp[le.__name__]:
                            temp[le.__name__]['metrices'] = []
                        else:
                            temp[le.__name__]['metrices'].append(val[0])

    print(temp)
    return temp

if __name__ == '__main__':
    eval(cmd ())
    
#/Users/suvodeepmajumder/Documents/AI4SE/magician_package/magician_package/data/preprocessed



from __future__ import print_function, division

__author__ = 'amrit'

import sys
from .demo import cmd
#sys.dont_write_bytecode = True
from collections import OrderedDict
import os
from random import seed, shuffle
import numpy as np
from collections import Counter
from .featurization import *
from .ML import DT, SVM, RF, FFT1
from sklearn.model_selection import StratifiedKFold
import pickle
import time
import warnings

def fxn():
    warnings.warn("deprecated", DeprecationWarning)

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    fxn()

ROOT = os.getcwd()
files = ["pitsA", "pitsB", "pitsC", "pitsD", "pitsE", "pitsF"]
MLS = [DT, RF, SVM,FFT1]
#MLS = [DT]
MLS_para_dic = {'DT':OrderedDict([("min_samples_split", 2), ("min_impurity_decrease", 0.0), ("max_depth", None),
                             ("min_samples_leaf", 1)]), 
                'RF': OrderedDict([("min_samples_split", 2),("max_leaf_nodes", None), ("min_samples_leaf", 1),
                                   ("min_impurity_decrease", 0.0),("n_estimators", 10)]),
                'SVM': OrderedDict([("C", 1.0), ("kernel", 'linear'),
                             ("degree", 3)]), 
                'FFT1':OrderedDict()}

metrics = ['accuracy', 'recall', 'precision', 'false_alarm']
#features = ['10', '25', '50', '100']
features = ['10']

def readfile1(filename=''):
    dict = []
    print(filename)
    labels = []
    with open(filename, 'r') as f:
        for doc in f.readlines():
            try:
                row = doc.lower().split(">>>")
                dict.append(row[0].strip())
                labels.append(row[1].strip())
            except:
                pass
    count = Counter(labels)
    import operator
    key = max(count.iteritems(), key=operator.itemgetter(1))[0]
    labels = map(lambda x: 1 if x == key else 0, labels)
    return np.array(dict), np.array(labels)


def _test(res='',_MLS = [DT, RF, SVM,FFT1],file_path = ''):
    seed(1)
    np.random.seed(1)
    #path = ROOT + "/data/preprocessed/" + res + ".txt"
    path = file_path + res + ".txt"
    raw_data, labels = readfile1(path)
    temp = {}
    MLS = _MLS
    for i in range(2):
        print('Running iteration ',i)
        ranges = range(len(labels))
        shuffle(ranges)
        raw_data = raw_data[ranges]
        labels = labels[ranges]

        for fea in features:
            if fea not in temp:
                temp[fea] = {}
            start_time = time.time()
            corpus, _ = LDA_(raw_data,n_components=int(fea))
            end_time = time.time() - start_time
            skf = StratifiedKFold(n_splits=2)
            for train_index, test_index in skf.split(corpus, labels):
                train_data, train_labels = corpus[train_index], labels[train_index]
                test_data, test_labels = corpus[test_index], labels[test_index]
                for j, le in enumerate(MLS):
                    if le.__name__ not in temp[fea]:
                        temp[fea][le.__name__] = {}
                    start_time1 = time.time()

                    _,val = MLS[j](MLS_para_dic[le.__name__], train_data, train_labels, test_data, test_labels, 'recall')
                    end_time1 = time.time() - start_time1
                    for m in metrics:
                        if m not in temp[fea][le.__name__]:
                            temp[fea][le.__name__][m] = []
                        temp[fea][le.__name__][m].append(val[0][m])
                    if 'times' not in temp[fea][le.__name__]:
                        temp[fea][le.__name__]['times']=[]
                    else:
                        temp[fea][le.__name__]['times'].append(end_time1 + end_time)
                    if 'features' not in temp[fea][le.__name__]:
                        temp[fea][le.__name__]['features']=[]
                    else:
                        temp[fea][le.__name__]['features'].append(val[1])
    return temp


if __name__ == '__main__':
    eval(cmd())



from __future__ import print_function, division

__author__ = 'amrit'

import sys

#sys.dont_write_bytecode = True
import pickle
import os
import matplotlib.pyplot as plt
import matplotlib.gridspec as gridspec
import matplotlib.text as mpl_text
import numpy as np

ROOT=os.getcwd()
MLS=["DT", "RF", "SVM",  "FFT1"]
#metrics=['accuracy','recall','precision','false_alarm']
metrics=['recall','precision']
files=["pitsA", "pitsB", "pitsC", "pitsD", "pitsE", "pitsF"]
features = ['10', '25', '50', '100']

class AnyObject(object):
    def __init__(self, text, color):
        self.my_text = text
        self.my_color = color

class AnyObjectHandler(object):
    def legend_artist(self, legend, orig_handle, fontsize, handlebox):
        x0, y0 = handlebox.xdescent, handlebox.ydescent
        width, height = handlebox.width, handlebox.height
        patch = mpl_text.Text(x=0, y=0, text=orig_handle.my_text, color=orig_handle.my_color, verticalalignment=u'baseline',
                                horizontalalignment=u'left', multialignment=None,
                                fontproperties=None, linespacing=None,
                                rotation_mode=None)
        handlebox.add_artist(patch)
        return patch

def dump_files(f='',prefix=''):
    for _, _, files in os.walk(ROOT + "/../dump/"):
        for file in files:
            if file.startswith(prefix+f):
                return file

def LDADE_FFT(LDADE,LDA):
    tuned_med = {}
    for f in files:
        for m in metrics:
            if m not in tuned_med:
                tuned_med[m] = {}
            if f not in tuned_med[m]:
                tuned_med[m][f]=[]
                tuned_med[m][f].append(['LDADE_FFT'] + LDADE[f]['FFT1'][m])
                tuned_med[m][f].append(['LDADE_SVM'] + LDADE[f]['SVM'][m])

            for x in features:
                    tuned_med[m][f].append([str(x) + '_FFT'] + LDA[f][x]['FFT1'][m])
    return tuned_med

def FFT_features(LDA):
    tuned_med = {}
    for f in files:
        for m in ['features']:
            if m not in tuned_med:
                tuned_med[m] = {}
            if f not in tuned_med[m]:
                tuned_med[m][f]=[]
            for x in ['10']:
                    tuned_med[m][f].append([str(x) + '_FFT'] + LDA[f][x]['FFT1'][m])
    return tuned_med

def LDADE_FFT_runtimes(LDADE):
    tuned_med = {}
    for f in files:
        for m in ['times']:
            if m not in tuned_med:
                tuned_med[m] = {}
            if f not in tuned_med[m]:
                tuned_med[m][f]=[]
                tuned_med[m][f].append(['LDADE_FFT'] + [np.median(LDADE[f]['FFT1'][m])])
                tuned_med[m][f].append(['LDADE_SVM'] + [np.median(LDADE[f]['SVM'][m])])

    return tuned_med

def SVM_FFT(LDA,untuned):
    tuned_med = {}
    for f in files:
        for m in metrics:
            if m not in tuned_med:
                tuned_med[m] = {}
            if f not in tuned_med[m]:
                tuned_med[m][f]=[]
                tuned_med[m][f].append(['TFIDF_SVM'] + untuned[f]['TFIDF']['SVM'][m])

            for x in features:
                    tuned_med[m][f].append([str(x) + '_FFT'] + LDA[f][x]['FFT1'][m])
    return tuned_med

def SVM_FFT_runtimes(LDA,untuned):
    tuned_med = {}
    for f in files:
        for m in ['times']:
            if m not in tuned_med:
                tuned_med[m] = {}
            if f not in tuned_med[m]:
                tuned_med[m][f]=[]
                tuned_med[m][f].append(['TFIDF_SVM'] + [np.median(untuned[f]['TFIDF']['SVM'][m])])

            for x in features:
                    tuned_med[m][f].append([str(x) + '_FFT'] + [np.median(LDA[f][x]['FFT1'][m])])
    return tuned_med

def draw(dic):
    font = {'size': 70}
    plt.rc('font', **font)
    paras = {'lines.linewidth': 70, 'legend.fontsize': 70, 'axes.labelsize': 80, 'legend.frameon': True,
                  'figure.autolayout': True,'axes.linewidth':8}
    plt.rcParams.update(paras)

    boxprops = dict(linewidth=9,color='black')
    colors=['red','green', 'blue', 'orange','cyan','purple']*6
    whiskerprops = dict(linewidth=5)
    medianprops = dict(linewidth=8, color='firebrick')
    #meanpointprops = dict(marker='D', markeredgecolor='black',markerfacecolor='firebrick',markersize=20)

    fig = plt.figure(figsize=(80, 60))
    outer = gridspec.GridSpec(1, 1, wspace=0.1, hspace=0.2)
    for i,a in enumerate([1]):
        inner = gridspec.GridSpecFromSubplotSpec(2, 1, subplot_spec=outer[i], wspace=0.05, hspace=0.0)
        for j,b in enumerate(dic.keys()):
            ax = plt.Subplot(fig, inner[j])
            temp=[item[1:] for sublist in dic[b].values() for item in sublist]

            bplot=ax.boxplot(temp,showmeans=False,showfliers=False,medianprops=medianprops,capprops=whiskerprops,
                       flierprops=whiskerprops,boxprops=boxprops,whiskerprops=whiskerprops,
                       positions=[1,2,3,4,5,6, 8,9,10,11,12,13 ,15,16,17,18,19,20 ,22,23,24,25,26,27 ,29,30,31,32,33,34,
                                  36,37,38,39,40,41])
            for patch, color in zip(bplot['boxes'], colors):
                patch.set(color=color)
            ax.set_xticks([3.5,11.5,17.5,24.5,31.5,38.5])
            ax.set_xticklabels(dic[b].keys())
            ax.set_ylabel(b,labelpad=30)
            #ax.set_ylim([0,1])
            if j!=1:
                plt.setp(ax.get_xticklabels(), visible=False)
            fig.add_subplot(ax)

    # box1 = TextArea("DT", textprops=dict(color=colors[0],size='large'))
    # box2 = TextArea("RF", textprops=dict(color=colors[1],size='large'))
    # box3 = TextArea("SVM", textprops=dict(color=colors[2],size='large'))
    # box = HPacker(children=[box1, box2, box3],
    #               align="center",
    #               pad=0, sep=5)
    #
    # anchored_box = AnchoredOffsetbox(loc=3,child=box, pad=0.,frameon=True,
    #                                  bbox_to_anchor=(0., 1.02),borderpad=0.)
    #
    # plt.artist(anchored_box)
    obj_0 = AnyObject("LDADE_FFT", colors[0])
    obj_1 = AnyObject("LDADE_SVM", colors[1])
    obj_2 = AnyObject("10_FFT", colors[2])
    obj_3 = AnyObject("25_FFT", colors[3])
    obj_4 = AnyObject("50_FFT", colors[4])
    obj_5 = AnyObject("100_FFT", colors[5])

    plt.legend([obj_0, obj_1,obj_2,obj_3,obj_4,obj_5], ['LDADE_FFT','LDADE_SVM', '10_FFT', '25_FFT','50_FFT','100_FFT'],
               handler_map={obj_0: AnyObjectHandler(), obj_1: AnyObjectHandler(),obj_2: AnyObjectHandler(),
               obj_3: AnyObjectHandler(), obj_4: AnyObjectHandler(),obj_5: AnyObjectHandler()},
               loc='upper center', bbox_to_anchor=(0.5, 2.1),
               fancybox=True, shadow=True, ncol=6,handletextpad=4)
    # plt.figtext(0.40, 0.9, 'DT', color=colors[0],size='large')
    # plt.figtext(0.50, 0.9, 'RF', color=colors[1],size='large')
    # plt.figtext(0.60, 0.9, 'SVM', color=colors[2],size='large')

    plt.savefig("../results/graph1.png", bbox_inches='tight')
    plt.close(fig)

def draw1(dic):
    font = {'size': 70}
    plt.rc('font', **font)
    paras = {'lines.linewidth': 70, 'legend.fontsize': 70, 'axes.labelsize': 80, 'legend.frameon': True,
                  'figure.autolayout': True,'axes.linewidth':8}
    plt.rcParams.update(paras)

    boxprops = dict(linewidth=9,color='black')
    colors=['red','green', 'blue', 'orange','purple']*6
    whiskerprops = dict(linewidth=5)
    medianprops = dict(linewidth=8, color='firebrick')
    #meanpointprops = dict(marker='D', markeredgecolor='black',markerfacecolor='firebrick',markersize=20)

    fig = plt.figure(figsize=(80, 60))
    outer = gridspec.GridSpec(1, 1, wspace=0.1, hspace=0.2)
    for i,a in enumerate([1]):
        inner = gridspec.GridSpecFromSubplotSpec(2, 1, subplot_spec=outer[i], wspace=0.05, hspace=0.0)
        for j,b in enumerate(dic.keys()):
            ax = plt.Subplot(fig, inner[j])
            temp=[item[1:] for sublist in dic[b].values() for item in sublist]

            bplot=ax.boxplot(temp,showmeans=False,showfliers=False,medianprops=medianprops,capprops=whiskerprops,
                       flierprops=whiskerprops,boxprops=boxprops,whiskerprops=whiskerprops,
                       positions=[1,2,3,4,5, 7,8,9,10,11, 13,14,15,16,17 ,19,20,21,22,23 ,25,26,27,28,29, 31,32,33,34,35
                                  ])
            for patch, color in zip(bplot['boxes'], colors):
                patch.set(color=color)
            ax.set_xticks([3,9,15,21,27,33])
            ax.set_xticklabels(dic[b].keys())
            ax.set_ylabel(b,labelpad=30)
            #ax.set_ylim([0,1])
            if j!=1:
                plt.setp(ax.get_xticklabels(), visible=False)
            fig.add_subplot(ax)

    # box1 = TextArea("DT", textprops=dict(color=colors[0],size='large'))
    # box2 = TextArea("RF", textprops=dict(color=colors[1],size='large'))
    # box3 = TextArea("SVM", textprops=dict(color=colors[2],size='large'))
    # box = HPacker(children=[box1, box2, box3],
    #               align="center",
    #               pad=0, sep=5)
    #
    # anchored_box = AnchoredOffsetbox(loc=3,child=box, pad=0.,frameon=True,
    #                                  bbox_to_anchor=(0., 1.02),borderpad=0.)
    #
    # plt.artist(anchored_box)
    obj_0 = AnyObject("TFIDF_SVM", colors[0])
    obj_1 = AnyObject("10_FFT", colors[1])
    obj_2 = AnyObject("25_FFT", colors[2])
    obj_3 = AnyObject("50_FFT", colors[3])
    obj_4 = AnyObject("100_FFT", colors[4])

    plt.legend([obj_0, obj_1,obj_2,obj_3,obj_4], ['TFIDF_SVM', '10_FFT', '25_FFT','50_FFT','100_FFT'],
               handler_map={obj_0: AnyObjectHandler(), obj_1: AnyObjectHandler(),obj_2: AnyObjectHandler(),
               obj_3: AnyObjectHandler(), obj_4: AnyObjectHandler()},
               loc='upper center', bbox_to_anchor=(0.5, 2.1),
               fancybox=True, shadow=True, ncol=5,handletextpad=4)
    # plt.figtext(0.40, 0.9, 'DT', color=colors[0],size='large')
    # plt.figtext(0.50, 0.9, 'RF', color=colors[1],size='large')
    # plt.figtext(0.60, 0.9, 'SVM', color=colors[2],size='large')

    plt.savefig("../results/graph2.png", bbox_inches='tight')
    plt.close(fig)

def for_LDADE():
    filenames=[]
    dic={}
    for f in files:
        filenames.append(dump_files(f,'LDADE'))
    for f in filenames:
        with open("../dump/" + f, 'rb') as handle:
            g=f.split(".pickle")[0].split("LDADE")[1]
            dic[g] = pickle.load(handle)
    return dic

def for_LDA():
    filenames=[]
    dic={}
    for f in files:
        filenames.append(dump_files(f,'LDA'))
    for f in filenames:
        with open("../dump/" + f, 'rb') as handle:
            g = f.split(".pickle")[0].split("LDA")[1]
            dic[g] = pickle.load(handle)
    return dic

def for_untuned():
    filenames=[]
    dic={}
    for f in files:
        filenames.append(dump_files(f,'untuned'))
    for f in filenames:
        with open("../dump/" + f, 'rb') as handle:
            g = f.split(".pickle")[0].split("untuned")[1]
            dic[g] = pickle.load(handle)
    return dic


if __name__ == '__main__':
    LDADE=for_LDADE()
    LDA=for_LDA()
    untuned = for_untuned()
    dic=LDADE_FFT(LDADE,LDA)
    ## RQ2
    print(dic)
    #draw(dic)

    dic=SVM_FFT(LDA,untuned)
    ## RQ1
    print(dic)
    #draw1(dic)

    # dic=SVM_FFT_runtimes(LDA,untuned)
    # print(dic)
    # dic=LDADE_FFT_runtimes(LDADE)
    # print(dic)

    # print(FFT_features(LDA))




from __future__ import print_function, division

__author__ = 'amrit'

import sys

#sys.dont_write_bytecode = True
from random import shuffle, seed
import numpy as np
import os
from sklearn.feature_extraction.text import CountVectorizer
import copy
from sklearn.decomposition import LatentDirichletAllocation

ROOT=os.getcwd()
seed(1)
np.random.seed(1)

def calculate(topics=[], lis=[], count1=0):
    count = 0
    for i in topics:
        if i in lis:
            count += 1
    if count >= count1:
        return count
    else:
        return 0


def recursion(topic=[], index=0, count1=0):
    count = 0
    global data
    # print(data)
    # print(topics)
    d = copy.deepcopy(data)
    d.pop(index)
    for l, m in enumerate(d):
        # print(m)
        for x, y in enumerate(m):
            if calculate(topics=topic, lis=y, count1=count1) != 0:
                count += 1
                break
                # data[index+l+1].pop(x)
    return count


data = []


def jaccard(a, score_topics=[], term=0):
    labels = []  # ,6,7,8,9]
    labels.append(term)
    global data
    l = []
    data = []
    file_data = {}
    for doc in score_topics:
        l.append(doc.split())
    for i in range(0, len(l), int(a)):
        l1 = []
        for j in range(int(a)):
            l1.append(l[i + j])
        data.append(l1)
    dic = {}
    for x in labels:
        j_score = []
        for i, j in enumerate(data):
            for l, m in enumerate(j):
                sum = recursion(topic=m, index=i, count1=x)
                if sum != 0:
                    j_score.append(sum / float(9))
                '''for m,n in enumerate(l):
                    if n in j[]'''
        dic[x] = j_score
        if len(dic[x]) == 0:
            dic[x] = [0]
    file_data['citemap'] = dic
    X = range(len(labels))
    Y_median = []
    Y_iqr = []
    for feature in labels:
        Y = file_data['citemap'][feature]
        Y = sorted(Y)
        return Y[int(len(Y) / 2)]


def get_top_words(model, path1, feature_names, n_top_words, i=0, file1=''):
    topics = []
    fo = open(path1, 'a+')
    fo.write("Run: " + str(i) + "\n")
    for topic_idx, topic in enumerate(model.components_):
        str1 = ''
        fo.write("Topic " + str(topic_idx) + ": ")
        for j in topic.argsort()[:-n_top_words - 1:-1]:
            str1 += feature_names[j] + " "
            fo.write(feature_names[j] + " ")
        str1=str(str1.encode('ascii', 'ignore'))
        topics.append(str1)
        fo.write("\n")
    fo.close()
    return topics


def readfile1(filename=''):
    dict = []
    with open(filename, 'r') as f:
        for doc in f.readlines():
            try:
                row = doc.lower().strip()
                dict.append(row)
            except:
                pass
    return dict


def _test_LDA( path1, file='', data_samples=[], term=7, **l):
    topics = []
    for i in range(10):
        shuffle(data_samples)

        tf_vectorizer = CountVectorizer(max_df=0.95, min_df=2, stop_words='english')
        tf = tf_vectorizer.fit_transform(data_samples)

        lda1 = LatentDirichletAllocation(max_iter=100,learning_method='online',**l)

        lda1.fit_transform(tf)

        # print("done in %0.3fs." % (time() - t0))
        tf_feature_names = tf_vectorizer.get_feature_names()
        topics.extend(get_top_words(lda1, path1, tf_feature_names, term, i=i, file1=file))
    return topics

## main(k,alpha,beta,tune='tuned',file='',term='',data_samples='')

def main(*x, **r):

    base = ROOT+"/../temp/"
    path = os.path.join(base, r['file'], str(r['term']))
    if not os.path.exists(path):
        os.makedirs(path)
    l = np.asarray(x)
    n_components = l[0]['n_components']
    doc_topic_prior=l[0]['doc_topic_prior']
    topic_word_prior=l[0]['topic_word_prior']
    path1 = path + "/K_" + str(n_components) + "_a_" + str(doc_topic_prior) + "_b_" + str(topic_word_prior) + ".txt"
    with open(path1, "w") as f:
        f.truncate()

    topics = _test_LDA(path1, file=r['file'],data_samples=r['data_samples'],term=int(r['term']),n_components=n_components,
                       doc_topic_prior=doc_topic_prior,topic_word_prior=topic_word_prior)

    a = jaccard(n_components, score_topics=topics, term=int(r['term']))
    fo = open(path1, 'a+')
    fo.write("\nScore: " + str(a))
    fo.close()
    return a, [{},[]]



import sys 
import os
#sys.path.insert(0,os.getcwd() + '/src')
from . import LDA
from . import LDADE
from .ML import DT, SVM, RF, FFT1
import warnings
from . import DE_ML
#first_arg = sys.argv[1]
#second_arg = sys.argv[2]


def fxn():
    warnings.warn("deprecated", DeprecationWarning)

with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    fxn()
warnings.filterwarnings("ignore", message="numpy.dtype size changed")
warnings.filterwarnings("ignore", message="numpy.ufunc size changed")

def _LDA(file_name = '',MLS = [DT, RF, SVM,FFT1],file_path = ''):
    print("The Program started============================================================= ")
    result_ = LDA._test(file_name,MLS,file_path)
    print("Program completed")
    return result_

def _LDADE(file_name = '',MLS = [DT, RF, SVM,FFT1],file_path = ''):
    print("The Program started============================================================= ")
    result_ = LDADE._test(file_name,MLS,file_path)
    print("Program completed")
    return result_

def learner(file_name,second_arg,file_path):
    MLS = []
    if second_arg == '1':
        MLS = [DT]
    elif second_arg == '2':
        MLS = [RF]
    elif second_arg == '3':
        MLS = [SVM]
    elif second_arg == '4':
        MLS = [FFT1]
    elif second_arg == '5':
        MLS = [DT, RF, SVM,FFT1]
    else:
        print('Invalid Argument')
        sys.exit()
    result_x = _LDA(file_name,MLS,file_path)
    return result_x

def semi_auto_learner(file_name,second_arg,file_path):
    MLS = []
    if second_arg == '1':
        MLS = [DT]
    elif second_arg == '2':
        MLS = [RF]
    elif second_arg == '3':
        MLS = [SVM]
    elif second_arg == '4':
        MLS = [FFT1]
    elif second_arg == '5':
        MLS = [DT, RF, SVM,FFT1]
    else:
        print('Invalid Argument')
        sys.exit()
    result_x = _LDADE(file_name,MLS,file_path)
    return result_x

def auto_learner(file_name,second_arg,file_path):
    MLS = []
    if second_arg == '1':
        MLS = [DT]
    elif second_arg == '2':
        MLS = [RF]
    elif second_arg == '3':
        MLS = [SVM]
    elif second_arg == '4':
        MLS = [DT, RF, SVM]
    result_x = DE_ML._test(file_name,MLS,file_path)
    return result_x


__author__ = 'amrit'

import sys

#sys.dont_write_bytecode = True
import collections
import numpy as np
import pandas as pd
from .helpers import get_performance, get_score, get_auc
from mdlp.discretization import MDLP

PRE, REC, SPEC, FPR, NPV, ACC, F1 = 7, 6, 5, 4, 3, 2, 1
MATRIX = "\t".join(["\tTP", "FP", "TN", "FN"])
PERFORMANCE = " \t".join(["\tCLF", "PRE ", "REC", "SPE", "FPR", "NPV", "ACC", "F_1"])


class FFT(object):

    def __init__(self, max_level=5):
        self.max_depth = max_level - 1
        cnt = 2 ** self.max_depth
        self.tree_cnt = cnt
        self.tree_depths = [0] * cnt
        self.best = -1
        self.target = "" # "bug"
        self.ignore = {} # moved to main. # {"name", "version", 'name.1', 'prediction'}
        self.split_method = "median"
        self.criteria = "Dist2Heaven"
        self.data_name = ''
        self.train, self.test = None, None
        self.structures = None
        self.computed_cache = {}
        self.selected = [{} for _ in range(cnt)]
        self.tree_scores = [None] * cnt
        self.dist2heavens = [None] * cnt
        self.node_descriptions = [None] * cnt
        self.performance_on_train = [collections.defaultdict(dict) for _ in xrange(cnt)]
        self.performance_on_test = [None] * cnt
        self.predictions = [None] * cnt
        self.loc_aucs = [None] * cnt
        self.print_enabled = False
        #self.results = {}

    "Get the loc_auc for a specific tree"
    def get_tree_loc_auc(self, data, i):
        if "loc" not in data:
            return 8
        # self.predict will add/modify the 'prediction' column to the data
        self.predict(data, i)
        sorted_data = data.sort_values(by=["prediction", "loc"], ascending=[False, True])
        return get_auc(sorted_data)

    "Build all possible tress."

    def build_trees(self):
        self.structures = self.get_all_structure()
        data = self.train
        for i in range(self.tree_cnt):
            self.grow(data, i, 0, [0, 0, 0, 0])
            #self.loc_aucs[i] = self.get_tree_loc_auc(data, i)


    "Evaluate all tress built on TEST data."

    def eval_trees(self):
        for i in range(self.tree_cnt):
            # Get performance on TEST data.
            self.eval_tree(i)

    "Evaluate the performance of the given tree on the TEST data."

    def eval_tree(self, t_id):
        if self.performance_on_test[t_id]:
            return
        depth = self.tree_depths[t_id]
        self.node_descriptions[t_id] = [[] for _ in range(depth + 1)]
        TP, FP, TN, FN = 0, 0, 0, 0
        data = self.test
        for level in range(depth + 1):
            cue, direction, threshold, decision = self.selected[t_id][level]
            undecided, metrics, loc_auc = self.eval_decision(data, cue, direction, threshold, decision)
            tp, fp, tn, fn = self.update_metrics(level, depth, decision, metrics)
            TP, FP, TN, FN = TP + tp, FP + fp, TN + tn, FN + fn
            if len(undecided) == 0:
                break
            data = undecided
        pre, rec, spec, fpr, npv, acc, f1 = get_performance([TP, FP, TN, FN])
        self.performance_on_test[t_id] = [TP, FP, TN, FN, pre, rec, spec, fpr, npv, acc, f1]

    "Find the best tree based on the score in TRAIN data."

    def find_best_tree(self):
        if self.tree_scores and self.tree_scores[0]:
            return
        if not self.performance_on_train or not self.performance_on_train[0]:
            self.grow()
        if self.print_enabled:
            print("\t----- PERFORMANCES FOR ALL FFTs on Training Data -----")
            print(PERFORMANCE + " \t" + self.criteria)
        best = [-1, float('inf')]
        for i in range(self.tree_cnt):
            all_metrics = self.performance_on_train[i][self.tree_depths[i]]
            if self.criteria == "LOC_AUC":
                score = self.loc_aucs[i]
            else:
                score = get_score(self.criteria, all_metrics[:4])
            self.tree_scores[i] = score
            self.dist2heavens[i] = get_score("Dist2Heaven", all_metrics[:4])
            if score < best[-1]:
                best = [i, score]
            if self.print_enabled:
                print("\t" + "\t".join(
                    ["FFT(" + str(i) + ")"] + \
                    [str(x).ljust(5, "0") for x in all_metrics[4:] + \
                     [score if self.criteria == "Dist2Heaven" else -score]]))
        if self.print_enabled:
            print("\tThe best tree found on training data is: FFT(" + str(best[0]) + ")")
        self.best = best[0]
        return best[0]

    "Given how the decision is made, get the description for the node."

    def describe_decision(self, t_id, level, metrics, reversed=False):
        cue, direction, threshold, decision = self.selected[t_id][level]
        tp, fp, tn, fn = metrics
        results = ["\'True\'", "\'False\'"]
        if type(threshold) != type(np.ndarray(1)):
            mappings = {">":"<=", "<":">="}
        else:
            mappings = {"inside":"outside", "outside":"inside"}
        direction = mappings[direction] if reversed else direction+" "
        description = ("\t| " * (level + 1) +
                       " ".join([str(cue), direction, str(threshold)]) +
                       "\t--> " + results[1 - decision if reversed else decision]).ljust(30, " ")
        pos = "\tFalse Alarm: " + str(fp) + ", Hit: " + str(tp)
        neg = "\tCorrect Rej: " + str(tn) + ", Miss: " + str(fn)
        if not reversed:
            description += pos if decision == 1 else neg
        else:
            description += neg if decision == 1 else pos
        return description

    "Given how the decision is made, get the performance for this decision."

    def eval_decision(self, data, cue, direction, threshold, decision):
        try:
            if type(threshold) != type(np.ndarray(1)):
                if direction == ">":
                    decided, undecided = data.loc[data[cue] > threshold], data.loc[data[cue] <= threshold]
                else:
                    decided, undecided = data.loc[data[cue] < threshold], data.loc[data[cue] >= threshold]
            else:
                decided = data.loc[(data[cue] >= threshold[0]) & (data[cue] < threshold[1])]
                undecided = data.loc[(data[cue] < threshold[0]) | (data[cue] >= threshold[1])]
        except:
            print("Exception")
            return 1, 2, 3
        if decision == 1:
            pos, neg = decided, undecided
        else:
            pos, neg = undecided, decided
        # get auc for loc.
        if "loc" in data:
            sorted_data = pd.concat([df.sort_values(by=["loc"], ascending=True) for df in [pos, neg]])
            loc_auc = get_auc(sorted_data)
        else:
            loc_auc = 0
        tp = pos.loc[pos[self.target] == 1]
        fp = pos.loc[pos[self.target] == 0]
        tn = neg.loc[neg[self.target] == 0]
        fn = neg.loc[neg[self.target] == 1]
        # pre, rec, spec, fpr, npv, acc, f1 = get_performance([tp, fp, tn, fn])
        # return undecided, [tp, fp, tn, fn, pre, rec, spec, fpr, npv, acc, f1]
        return undecided, map(len, [tp, fp, tn, fn]), loc_auc

    "Given data and the specific tree id, add a 'prediction' column to the dataframe."

    def predict(self, data, t_id=-1):
        # predictions = pd.Series([None] * len(data))
        if t_id == -1:
            t_id = self.best
        original = data
        original['prediction'] = pd.Series([None] * len(data))
        depth = self.tree_depths[t_id]
        for level in range(depth + 1):
            cue, direction, threshold, decision = self.selected[t_id][level]
            undecided, metrics, loc_auc = self.eval_decision(data, cue, direction, threshold, decision)
            decided_idx = [i for i in data.index if i not in undecided.index]
            # original['prediction'][decided_idx] = decision
            original.loc[decided_idx, 'prediction'] = decision
            data = undecided
        original.loc[data.index, 'prediction'] = 1 if decision == 0 else 0
        # original['prediction'][undecided.index] = 1 if decision == 0 else 0
        if None in original['prediction']:
            print("ERROR!")
        self.predictions[t_id] = original['prediction'].values
        return self.predictions[t_id]

    def eval_point_split(self, level, cur_selected, cur_performance, data, cue, direction, threshold, decision):
        TP, FP, TN, FN = cur_performance
        undecided, metrics, loc_auc = self.eval_decision(data, cue, direction, threshold, decision)
        tp, fp, tn, fn = self.update_metrics(level, self.max_depth, decision, metrics)
        # if the decision lead to no data, punish the score
        if sum([tp, fp, tn, fn]) == 0:
            score = float('inf')
        elif self.criteria == "LOC_AUC":
            score = loc_auc
        else:
            score = get_score(self.criteria, [TP + tp, FP + fp, TN + tn, FN + fn])
        if not cur_selected or score < cur_selected['score']:
            cur_selected = {'rule': (cue, direction, threshold, decision),\
                            'undecided': undecided,\
                            'metrics': [TP + tp, FP + fp, TN + tn, FN + fn], \
                            'score': score}
        return cur_selected


    def eval_range_split(self, level, cur_selected, cur_performance, data, cue, indexes, interval, decision):
        TP, FP, TN, FN = cur_performance
        pos = data.iloc[indexes]
        neg = data.iloc[~data.index.isin(indexes)]
        if decision == 1:
            decided = pos
            undecided = neg
        else:
            pos, neg = neg, pos
            decided = neg
            undecided = pos
        # get auc for loc.
        if "loc" in data:
            sorted_data = pd.concat([df.sort_values(by=["loc"], ascending=True) for df in [pos, neg]])
            loc_auc = get_auc(sorted_data)
        else:
            loc_auc = 0
        tp = pos.loc[pos[self.target] == 1]
        fp = pos.loc[pos[self.target] == 0]
        tn = neg.loc[neg[self.target] == 0]
        fn = neg.loc[neg[self.target] == 1]
        metrics = map(len, [tp, fp, tn, fn])
        tp, fp, tn, fn = self.update_metrics(level, self.max_depth, decision, metrics)
        # if the decision lead to no data, punish the score
        if sum([tp, fp, tn, fn]) == 0:
            score = float('inf')
        elif self.criteria == "LOC_AUC":
            score = loc_auc
        else:
            score = get_score(self.criteria, [TP + tp, FP + fp, TN + tn, FN + fn])
        if not cur_selected or score < cur_selected['score']:
            direction = "inside" if decision else "outside"
            cur_selected = {'rule': (cue, direction, interval, decision),\
                            'undecided': undecided,\
                            'metrics': [TP + tp, FP + fp, TN + tn, FN + fn],\
                            'score': score}
        return cur_selected


    "Grow the t_id_th tree for the level with the given data. Also save its performance on the TRAIN data"

    def grow(self, data, t_id, level, cur_performance):
        """
        :param data: current data for future tree growth
        :param t_id: tree id
        :param level: level id
        :return: None
        """
        if level >= self.max_depth:
            return
        if len(data) == 0:
            print("?????????????????????? Early Ends ???????????????????????")
            return
        self.tree_depths[t_id] = level
        decision = self.structures[t_id][level]
        structure = tuple(self.structures[t_id][:level + 1])
        cur_selected = self.computed_cache.get(structure, None)
        Y = data.as_matrix(columns=[self.target])
        if not cur_selected:
            for cue in list(data):
                if cue in self.ignore or cue == self.target:
                    continue
                if self.split_method == "MDLP":
                    mdlp = MDLP()
                    X = data.as_matrix(columns=[cue])
                    X_disc = mdlp.fit_transform(X, Y)
                    X_interval = np.asarray(mdlp.cat2intervals(X_disc, 0))
                    bins = np.unique(X_disc, axis=0)
                    if len(bins) <= 1:   # MDLP return the whole range as one bin, use median instead.
                        threshold = data[cue].median()
                        for direction in "><":
                            cur_selected = self.eval_point_split(level, cur_selected, cur_performance, data, cue, direction, threshold, decision)
                        continue
                    # print ", ".join([cue, str(bins)+" bins"])
                    for bin in bins:
                        indexes = np.where(X_disc == bin)[0]
                        interval = X_interval[indexes]
                        try:
                            if len(np.unique(interval, axis=0)) != 1:
                                print("???????????????????????????????????????????????????")
                        except:
                            print('ha')
                        interval = interval[0]
                        if interval[0] == float('-inf'):
                            threshold = interval[1]
                            for direction in "><":
                                cur_selected = self.eval_point_split(level, cur_selected, cur_performance, data, cue, direction, threshold, decision)
                        elif interval[1] == float('inf'):
                            threshold = interval[0]
                            for direction in "><":
                                cur_selected = self.eval_point_split(level, cur_selected, cur_performance, data, cue, direction, threshold, decision)
                        else:
                            cur_selected = self.eval_range_split(level, cur_selected, cur_performance, data, cue, indexes, interval, decision)
                    continue
                elif self.split_method == "percentile":
                    thresholds = set(data[cue].quantile([x/20.0 for x in range(1, 20)], interpolation='midpoint'))
                else:
                    thresholds = [data[cue].median()]
                # point split, e.g. median or x% percentiles.
                for threshold in thresholds:
                    for direction in "><":
                        cur_selected = self.eval_point_split(level, cur_selected, cur_performance, data, cue, direction, threshold, decision)

            self.computed_cache[structure] = cur_selected
        self.selected[t_id][level] = cur_selected['rule']
        self.performance_on_train[t_id][level] = cur_selected['metrics'] + get_performance(cur_selected['metrics'])
        self.grow(cur_selected['undecided'], t_id, level + 1, cur_selected['metrics'])

    # "describe a tree on the TEST data"
    #
    # def describe_tree(self, t_id):
    #     data = self.test
    #     depth = self.tree_depths[t_id]
    #     for level in range(depth + 1):
    #         cue, direction, threshold, decision = self.selected[t_id][level]
    #         undecided, metrics, loc_auc = self.eval_decision(data, cue, direction, threshold, decision)
    #         tp, fp, tn, fn = self.update_metrics(level, depth, decision, metrics)
    #         description = self.describe_decision(t_id, level, metrics)
    #         self.node_descriptions[t_id][level] += [description]
    #         if len(undecided) == 0:
    #             break
    #         data = undecided
    #     description = self.describe_decision(t_id, level, metrics, reversed=True)
    #     self.node_descriptions[t_id][level] += [description]

    "Given tree id, print the specific tree and its performances on the test data."

    def print_tree(self, t_id):
        data = self.test
        depth = self.tree_depths[t_id]
        string=''
        if not self.node_descriptions[t_id]:
            self.node_descriptions[t_id] = [[] for _ in range(depth + 1)]
        for i in range(depth + 1):
            if self.node_descriptions[t_id][i]:
                #print self.node_descriptions[t_id][i][0]
                string+=self.node_descriptions[t_id][i][0]+'\n'
            else:
                cue, direction, threshold, decision = self.selected[t_id][i]
                undecided, metrics, loc_auc = self.eval_decision(data, cue, direction, threshold, decision)
                description = self.describe_decision(t_id, i, metrics)
                self.node_descriptions[t_id][i] += [description]
                #print description
                string+=description+'\n'
                if len(undecided) == 0:
                    break
                data = undecided
        return string
        # description = self.describe_decision(t_id, i, metrics, reversed=True)
        # self.node_descriptions[t_id][i] += [description]
        # dist2heaven = get_score("Dist2Heaven", self.performance_on_test[t_id][:4])
        # loc_auc = -self.get_tree_loc_auc(self.test, t_id)
        #
        # if self.print_enabled:
        #     print self.node_descriptions[t_id][i][1]
        #     print "\t----- CONFUSION MATRIX -----"
        #     print MATRIX
        #     print "\t" + "\t".join(map(str, self.performance_on_test[t_id][:4]))
        #
        #     print "\t----- PERFORMANCES ON TEST DATA -----"
        #     print PERFORMANCE + " \tD2H"+ " \tLOC"
        #     print "\t" + "\t".join(
        #         ["FFT(" + str(self.best) + ")"] + \
        #         [str(x).ljust(5, "0") for x in self.performance_on_test[t_id][4:11] + [dist2heaven, loc_auc]])
        #         # map(str, ["FFT(" + str(self.best) + ")"] + self.performance_on_test[t_id][4:] + [dist2heaven]))

    "Get all possible tree structure"

    def get_all_structure(self):
        def dfs(cur, n):
            if len(cur) == n:
                ans.append(cur)
                return
            dfs(cur + [1], n)
            dfs(cur + [0], n)

        if self.max_depth < 0:
            return []
        ans = []
        dfs([], self.max_depth)
        return ans

    "Update the metrics(TP, FP, TN, FN) based on the decision."

    def update_metrics(self, level, depth, decision, metrics):
        tp, fp, tn, fn = metrics
        if level < depth:  # Except the last level, only part of the data(pos or neg) is decided.
            if decision == 1:
                tn, fn = 0, 0
            else:
                tp, fp = 0, 0
        return tp, fp, tn, fn



__author__ = 'amrit'

import sys

#sys.dont_write_bytecode = True
"""Helper functions."""
import math
import pickle
import numpy as np
import pandas as pd
from sklearn.metrics import auc

PRE, REC, SPEC, FPR, NPV, ACC, F1 = 7, 6, 5, 4, 3, 2, 1


def save_obj(obj, path):
    with open(path, 'wb') as f:
        pickle.dump(obj, f, pickle.HIGHEST_PROTOCOL)


def load_obj(path):
    with open(path, 'rb') as f:
        return pickle.load(f)


"Given prediction and truth, get tp, fp, tn, fn. "


def get_abcd(predict, truth):
    # pos > 0, neg == 0
    n = len(predict)
    tp, fp, tn, fn = 0, 0, 0, 0
    for i in range(n):
        if predict[i] > 0 and truth[i] > 0:
            tp += 1
        elif predict[i] > 0 and truth[i] == 0:
            fp += 1
        elif predict[i] == 0 and truth[i] == 0:
            tn += 1
        elif predict[i] == 0 and truth[i] > 0:
            fn += 1
    return tp, fp, tn, fn


"Given TP, FP, TN, FN, get all the other metrics. "


def get_performance(metrics):
    tp, fp, tn, fn = metrics
    pre = 1.0 * tp / (tp + fp) if (tp + fp) != 0 else 0
    rec = 1.0 * tp / (tp + fn) if (tp + fn) != 0 else 0
    spec = 1.0 * tn / (tn + fp) if (tn + fp) != 0 else 0
    fpr = 1 - spec
    npv = 1.0 * tn / (tn + fn) if (tn + fn) != 0 else 0
    acc = 1.0 * (tp + tn) / (tp + tn + fp + fn) if (tp + tn + fp + fn) != 0 else 0
    f1 = 2.0 * tp / (2.0 * tp + fp + fn) if (2.0 * tp + fp + fn) != 0 else 0
    return [round(x, 3) for x in [pre, rec, spec, fpr, npv, acc, f1]]


"Given the general metrics, return the score got by the specific criteria."


def get_score(criteria, metrics):   # The smaller the better
    tp, fp, tn, fn = metrics
    pre, rec, spec, fpr, npv, acc, f1 = get_performance([tp, fp, tn, fn])
    all_metrics = [tp, fp, tn, fn, pre, rec, spec, fpr, npv, acc, f1]
    if criteria == "accuracy":
        score = -all_metrics[-ACC]
    elif criteria == "Dist2Heaven":
        score = all_metrics[-FPR] ** 2 + (1 - all_metrics[-REC]) ** 2
        score = math.sqrt(score) / math.sqrt(2)
    elif criteria == "Gini":
        p1 = all_metrics[-PRE]  # target == 1 for the positive split
        p0 = 1 - all_metrics[-NPV]  # target == 1 for the negative split
        score = 1 - p0 ** 2 - p1 ** 2
    elif criteria== "precision":
        score=-all_metrics[-PRE]
    elif criteria== "recall":
        score=-all_metrics[-REC]
    elif criteria == "false_alarm":
        score = all_metrics[-FPR]
    else:  # Information Gain
        P, N = all_metrics[0] + all_metrics[3], all_metrics[1] + all_metrics[2]
        p = 1.0 * P / (P + N) if P + N > 0 else 0  # before the split
        p1 = all_metrics[-PRE]  # the positive part of the split
        p0 = 1 - all_metrics[-NPV]  # the negative part of the split
        I, I0, I1 = (-x * np.log2(x) if x != 0 else 0 for x in (p, p0, p1))
        I01 = p * I1 + (1 - p) * I0
        score = -(I - I01)  # the smaller the better.
    return round(score, 3)


def subtotal(x):
    xx = [0]
    for i, t in enumerate(x):
        xx += [xx[-1] + t]
    return xx[1:]


def get_recall(true):
    total_true = float(len([i for i in true if i == 1]))
    hit = 0.0
    recall = []
    for i in xrange(len(true)):
        if true[i] == 1:
            hit += 1
        recall += [hit / total_true if total_true else 0.0]
    return recall


def get_auc(data):
    """The smaller the better"""
    if len(data) == 1:
        return 0
    x_sum = float(sum(data['loc']))
    x = data['loc'].apply(lambda t: t / x_sum)
    xx = subtotal(x)
    yy = get_recall(data['bug'].values)
    try:
        ret = -round(auc(xx, yy), 3)
    except:
        print("?")
        ret = 0
    return ret


